#ifndef K3DUI_CHOOSER_H
#define K3DUI_CHOOSER_H

// K-3D
// Copyright (c) 1995-2004, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares k3d::chooser classes, which provide a model-view-controller UI for enumerated (fixed set of choices) data
		\author Tim Shead (tshead@k-3d.com)
*/

#include "k3dcontrol.h"

#include <k3dsdk/ienumeration_property.h>
#include <k3dsdk/iproperty.h>
#include <k3dsdk/iwritable_property.h>
#include <k3dsdk/result.h>
#include <k3dsdk/string_modifiers.h>

namespace k3d
{

namespace chooser
{

/////////////////////////////////////////////////////////////////////////////
// idata_proxy

/// Abstract interface for an object that proxies a data source for a chooser control (i.e. the "model" in model-view-controller)
class idata_proxy
{
public:
	virtual ~idata_proxy() {}

	/// Called to return the underlying data value
	virtual const std::string value() = 0;
	/// Called to set a new data value
	virtual void set_value(const std::string& Value) = 0;
	/// Signal emitted if the underlying data changes
	typedef SigC::Signal0<void> changed_signal_t;
	/// Signal emitted if the underlying data changes
	virtual changed_signal_t& changed_signal() = 0;

protected:
	idata_proxy() {}
	idata_proxy(const idata_proxy& RHS) {}
	idata_proxy& operator=(const idata_proxy& RHS) { return *this; }
};

/////////////////////////////////////////////////////////////////////////////
// control

/// Provides a UI for manipulating enumerated quantities (i.e. the view and the controller from model-view-controller)
class control :
	public k3dControl
{
public:
	control(k3d::iunknown* const CommandNodeParent, const std::string& CommandNodeName);
	control(k3d::iunknown* const CommandNodeParent, const std::string& CommandNodeName, const ienumeration_property::values_t& Choices);
	~control();

	/// Called by the framework when instantiating the button from a GTKML (XML) document
	bool Create(sdpGtkIObjectContainer* const ObjectContainer, sdpxml::Document& Document, sdpxml::Element& Element);

	/// Attaches the button to the data it fronts for
	bool attach(std::auto_ptr<idata_proxy> Data, k3d::istate_recorder* const StateRecorder, const std::string StateChangeName);

	const std::string CustomType() const;

	// k3d::icommand_node implementation
	bool execute_command(const std::string& Command, const std::string& Arguments);

private:
	typedef k3dControl base;

	/// Called whenever the the button state needs to be updated
	void update();

	/// Called to handle SDPGTK events
	void OnEvent(sdpGtkEvent* Event);
	/// Called when the GTK+ widgets are about to disappear
	void on_destroy();
	void on_value(sdpGtkEvent* Event);
	void on_value_toggled(sdpGtkEvent* Event);

	/// Store a reference to the underlying data object
	std::auto_ptr<idata_proxy> m_data;

	ienumeration_property::values_t m_choices;
};

/// Provides an implementation of k3d::chooser::idata_proxy that supports any data source that supports the value(), set_value(), and changed_signal() concepts
template<typename data_t>
class data_proxy :
	public idata_proxy
{
public:
	data_proxy(data_t& Data) :
		m_data(Data)
	{
	}

	const std::string value()
	{
		return k3d::to_string(m_data.value());
	}

	void set_value(const std::string& Value)
	{
		m_data.set_value(k3d::from_string(Value, static_cast<typename data_t::value_t>(0)));
	}

	changed_signal_t& changed_signal()
	{
		return m_data.changed_signal();
	}

private:
	data_proxy(const data_proxy& RHS);
	data_proxy& operator=(const data_proxy& RHS);
	~data_proxy() {}

	data_t& m_data;
};

/// Specialization of k3d::chooser::data_proxy for use with k3d::iproperty objects
template<>
class data_proxy<iproperty> :
	public idata_proxy
{
public:
	typedef iproperty data_t;

	data_proxy(data_t& Data) :
		m_readable_data(Data),
		m_writable_data(dynamic_cast<iwritable_property*>(&Data))
	{
	}

	const std::string value()
	{
		return boost::any_cast<std::string>(m_readable_data.value());
	}

	void set_value(const std::string& Value)
	{
		return_if_fail(m_writable_data);
		m_writable_data->set_value(Value);
	}

	changed_signal_t& changed_signal()
	{
		return m_readable_data.changed_signal();
	}

private:
	data_proxy(const data_proxy& RHS);
	data_proxy& operator=(const data_proxy& RHS);
	~data_proxy() {}

	data_t& m_readable_data;
	iwritable_property* const m_writable_data;
};

/// Convenience factory function for creating chooser::idata_proxy objects
template<typename data_t>
std::auto_ptr<idata_proxy> proxy(data_t& Data)
{
	return std::auto_ptr<idata_proxy>(new data_proxy<data_t>(Data));
}

} // namespace chooser

} // namespace k3d

#endif // !K3DUI_CHOOSER_H


